#!/usr/bin/env bash

export PROJECT_NAME={PROJECT_NAME}
export PROJECT_SHELL_SCRIPT={PROJECT_SHELL_SCRIPT}
export BRANCH_NAME={BRANCH_NAME}
export THROUGHPUT_GBPS={THROUGHPUT_GBPS}
export INSTANCE_TYPE={INSTANCE_TYPE}
export REGION={REGION}
export USER_NAME={USER_NAME}
export RUN_PROJECT_LOG_FN={RUN_PROJECT_LOG_FN}
export PUBLISH_METRICS_LOG_FN={PUBLISH_METRICS_LOG_FN}
export RUN_COMMAND={RUN_COMMAND}
export TEST_OBJECT_NAME={TEST_OBJECT_NAME}
export S3_BUCKET_NAME={S3_BUCKET_NAME}

export USER_DIR=/home/$USER_NAME/

$PROJECT_SHELL_SCRIPT "$RUN_COMMAND" 2>&1 $RUN_PROJECT_LOG_FN
